/*
    This is a virtual disk driver for Windows that uses one or more files to
    emulate physical disks.
    Copyright (C) 1999-2009 Bo Brantn.
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <windows.h>
#include <winioctl.h>
#include <shlobj.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "filedisk.h"

typedef	unsigned __int64 QWORD;

typedef	struct
	{
	QWORD	qwStart;
	QWORD	qwSize;
	CHAR 	cType[8];
	CHAR 	cLocation[16];
	}PartitionEntry;


#pragma pack(1)

typedef struct
	{
	BYTE	bBootIndicator;       		// 0x80 for bootable, 0 otherwise
    BYTE	bStartHead;           		// 0 based
    BYTE	bStartSector;         		// 1 based, bits 0..5, bits 6,7 are bits 8,9 of StartTrack
    BYTE	bStartTrack;          		// 0 based, bits 0..7
    BYTE	bOSType;              		// FAT-12: 1, FAT-16: 4, 6, 14, FAT-32: 11, 12
    BYTE	bEndHead;             		// see StartHead
    BYTE	bEndSector;           		// see StartSector
    BYTE	bEndTrack;					// see StartTrack
	DWORD	dwRelativeSector;			// offset to first sector of partition data
    DWORD	dwSectors;					// number of sectors in partition
	}PartitionRecord;

typedef struct
	{
	BYTE			bBootCode[440];
	DWORD			dwSerialNumber;
	WORD			wFree;
	PartitionRecord sTable[4];
	WORD			wSignature;			// 0xAA55
	}MasterBootRecord;


#pragma pack()


int ScanMBR(LPCTSTR pFilename,PartitionEntry *pEntries,int iMax)
{
	MasterBootRecord	sMbr;
	HANDLE				hFile;
	CHAR				cText[4];
	char 				cName[32*16];
	QWORD				qwNext[32];
	QWORD				qwTemp[32];
	QWORD				qwStart;
	QWORD				qwSize;
	QWORD				qwPos;
	DWORD				dwHigh;
	DWORD				dwBytes;
	int					iNum;
	int					iCnt;
	int					i,j;



	hFile = CreateFile(pFilename,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);

	if(hFile==INVALID_HANDLE_VALUE)
	{
		return -1;
	}


	iNum		= 1;
	iCnt		= 0;
	cName [0]	= 0;
	qwNext[0]	= 0;
	qwTemp[0]	= 0;
	qwSize		= GetFileSize(hFile,&dwHigh);
	qwSize     += (QWORD)dwHigh<<32;



	for(j=0;j<iNum && j<32;j++)
	{
		qwPos	= qwNext[j];
		dwHigh	= (DWORD)(qwPos>>32);

		SetFilePointer(hFile,(DWORD)qwPos,(LPLONG)&dwHigh,FILE_BEGIN);

		if(!ReadFile(hFile,&sMbr,512,&dwBytes,NULL))
		{
			break;
		}

		if(sMbr.wSignature!=0xAA55 ||						// is it a MBR
		  (sMbr.sTable[0].bBootIndicator&0x7F) ||
		  (sMbr.sTable[1].bBootIndicator&0x7F) ||
		  (sMbr.sTable[2].bBootIndicator&0x7F) ||
		  (sMbr.sTable[3].bBootIndicator&0x7F))
		{
			if(!qwTemp[j] || qwNext[j]==qwTemp[j])continue;
			qwNext[j]=qwTemp[j];
			qwTemp[j]=0;
			j--;
			continue;
		}


		for(i=0;i<4 && iNum<32;i++)							// scann MBR entries
		{
			if(!sMbr.sTable[i].dwRelativeSector)continue;
			if(!sMbr.sTable[i].dwSectors       )continue;

			qwStart   = sMbr.sTable[i].dwRelativeSector;
			qwStart  *= 512;
			qwStart  += qwPos;


			if(iCnt>=iMax)break;

			pEntries->qwStart	= qwStart;
			pEntries->qwSize	= sMbr.sTable[i].dwSectors;
			pEntries->qwSize   *= 512;

			if(cName[j*16])
				{
				cText[0] = '.';
				cText[1] = '1'+i;
				cText[2] = 0;
				strcpy(pEntries->cLocation,cName+j*16);
				strcat(pEntries->cLocation,cText);
				}
			else{
				pEntries->cLocation[0] = '1'+i;
				pEntries->cLocation[1] = 0;
				}

															// extendet patition
			if(sMbr.sTable[i].bOSType==0x05 ||sMbr.sTable[i].bOSType==0x0F)
			{
				strcpy(cName+iNum*16,pEntries->cLocation);
				qwNext[iNum] = qwStart;
				qwTemp[iNum] = 0;

				if(i>0)										// alternative position
				{
					qwTemp[iNum]   = sMbr.sTable[i-1].dwRelativeSector;
					qwTemp[iNum]  += sMbr.sTable[i-1].dwSectors;
					qwTemp[iNum]  *= 512;
					qwTemp[iNum]  += qwPos;
				}

				iNum++;
			}
			else
			{
				strcpy(pEntries->cType,"???");

				if(sMbr.sTable[i].bOSType==0x07 || // NTFS Partitionen
				   sMbr.sTable[i].bOSType==0x17)
					{
					strcpy(pEntries->cType,"NTFS");
					}

				if(sMbr.sTable[i].bOSType==0x01 ||	// FAT Partitionen
				   sMbr.sTable[i].bOSType==0x04 ||
				   sMbr.sTable[i].bOSType==0x06 ||
				   sMbr.sTable[i].bOSType==0x0B ||
				   sMbr.sTable[i].bOSType==0x0C ||
				   sMbr.sTable[i].bOSType==0x0E ||
				   sMbr.sTable[i].bOSType==0x11 ||
				   sMbr.sTable[i].bOSType==0x14 ||
				   sMbr.sTable[i].bOSType==0x16 ||
				   sMbr.sTable[i].bOSType==0x1B ||
				   sMbr.sTable[i].bOSType==0x1C ||
				   sMbr.sTable[i].bOSType==0x1E)
					{
					strcpy(pEntries->cType,"FAT");
					}

				pEntries++;
				iCnt++;
			}
		}
	}



	CloseHandle(hFile);


return iCnt;
}

int FileDiskSyntax(void)
{
    fprintf(stderr, "syntax:\n");
    fprintf(stderr, "filedisk /mount  <devicenumber> <filename> [size[k|M|G] | /ro | /cd | /p?] <drive:>\n");
	fprintf(stderr, "filedisk /list   <filename>\n");
    fprintf(stderr, "filedisk /umount <drive:>\n");
    fprintf(stderr, "filedisk /status <drive:>\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "filename formats:\n");
    fprintf(stderr, "  c:\\path\\filedisk.img\n");
    fprintf(stderr, "  \\Device\\Harddisk0\\Partition1\\path\\filedisk.img\n");
    fprintf(stderr, "  \\\\server\\share\\path\\filedisk.img\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "example:\n");
    fprintf(stderr, "filedisk /mount  0 c:\\tmp\\filedisk.img 8M f:\n");
    fprintf(stderr, "filedisk /mount  1 c:\\tmp\\cdimage.iso /cd i:\n");
    fprintf(stderr, "filedisk /umount f:\n");
    fprintf(stderr, "filedisk /umount i:\n");

    return -1;
}

void PrintLastError(char* Prefix)
{
    LPVOID lpMsgBuf;

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        GetLastError(),
        0,
        (LPTSTR) &lpMsgBuf,
        0,
        NULL
        );

    fprintf(stderr, "%s %s", Prefix, (LPTSTR) lpMsgBuf);

    LocalFree(lpMsgBuf);
}

int FileDiskListPartitions(LPTSTR pFilename)
{
	PartitionEntry	aEntries[16];
	QWORD			qwSize;
	int				iCount;
	CHAR			cSign;
	int				iPos;

	iCount = ScanMBR(pFilename,aEntries,16);

	if(iCount<0)
	{
        PrintLastError(pFilename);
        return -1;
	}


	if(iCount==0)
	{
        printf("\nno parition found !\n\n");
        return -1;
	}

	printf(" num          start      size   type   location   \n");
	printf("-------------------------------------------------\n");


	for(iPos=0;iPos<iCount;iPos++)
	{
		qwSize  = aEntries[iPos].qwSize;
		qwSize += 1023;
		qwSize /= 1024;

		if(qwSize<10000)
		{
			cSign='K';
		}
		else
		{
			qwSize += 1023;
			qwSize /= 1024;

			if(qwSize<10000)
			{
				cSign='M';
			}
			else
			{
				qwSize /= 1024;
				cSign='G';
			}
		}

		printf(" %2i: %14I64u  %6I64u %cB  %-6s %s\n",iPos+1,aEntries[iPos].qwStart,qwSize,cSign,aEntries[iPos].cType,aEntries[iPos].cLocation);
	}

	printf("-------------------------------------------------\n\n");



	return 0;
}


int
FileDiskMount(
    int						   DeviceNumber,
    POPEN_FILE_INFORMATION_EX  OpenFileInformation,
    BOOLEAN					   CdImage
)
{
    char    		VolumeName[] = "\\\\.\\ :";
    char    		DriveName[] = " :\\";
    char    		DeviceName[255];
    HANDLE			Device;
    DWORD			BytesReturned;




    VolumeName[4] = OpenFileInformation->DriveLetter;
    DriveName [0] = OpenFileInformation->DriveLetter;

    Device = CreateFile(
        VolumeName,
        GENERIC_READ | GENERIC_WRITE,
        FILE_SHARE_READ | FILE_SHARE_WRITE,
        NULL,
        OPEN_EXISTING,
        FILE_FLAG_NO_BUFFERING,
        NULL
        );

    if (Device != INVALID_HANDLE_VALUE)
    {
        CloseHandle(Device);
        SetLastError(ERROR_BUSY);
        PrintLastError(&VolumeName[4]);
        return -1;
    }

    if (CdImage)
    {
        sprintf(DeviceName, DEVICE_NAME_PREFIX "Cd" "%u", DeviceNumber);
    }
    else
    {
        sprintf(DeviceName, DEVICE_NAME_PREFIX "%u", DeviceNumber);
    }

    if (!DefineDosDevice(
        DDD_RAW_TARGET_PATH,
        &VolumeName[4],
        DeviceName
        ))
    {
        PrintLastError(&VolumeName[4]);
        return -1;
    }

    Device = CreateFile(
        VolumeName,
        GENERIC_READ | GENERIC_WRITE,
        FILE_SHARE_READ | FILE_SHARE_WRITE,
        NULL,
        OPEN_EXISTING,
        FILE_FLAG_NO_BUFFERING,
        NULL
        );

    if (Device == INVALID_HANDLE_VALUE)
    {
        PrintLastError(&VolumeName[4]);
        DefineDosDevice(DDD_REMOVE_DEFINITION, &VolumeName[4], NULL);
        return -1;
    }


	if(OpenFileInformation->FileOffset.QuadPart)
	{
		if (!DeviceIoControl(
			Device,
			IOCTL_FILE_DISK_OPEN_FILE_EX,
			OpenFileInformation,
			sizeof(OPEN_FILE_INFORMATION_EX) + OpenFileInformation->FileNameLength - 1,
			NULL,
			0,
			&BytesReturned,
			NULL
			))
		{
			PrintLastError("FileDisk:");
			DefineDosDevice(DDD_REMOVE_DEFINITION, &VolumeName[4], NULL);
			CloseHandle(Device);
			return -1;
		}
	}
	else
	{
		if (!DeviceIoControl(
			Device,
			IOCTL_FILE_DISK_OPEN_FILE,
			(POPEN_FILE_INFORMATION)&OpenFileInformation->FileSize,
			sizeof(OPEN_FILE_INFORMATION) + OpenFileInformation->FileNameLength - 1,
			NULL,
			0,
			&BytesReturned,
			NULL
			))
		{
			PrintLastError("FileDisk:");
			DefineDosDevice(DDD_REMOVE_DEFINITION, &VolumeName[4], NULL);
			CloseHandle(Device);
			return -1;
		}
	}

    CloseHandle(Device);

    SHChangeNotify(SHCNE_DRIVEADD, SHCNF_PATH, DriveName, NULL);

    return 0;
}

int FileDiskUmount(char DriveLetter)
{
    char    VolumeName[] = "\\\\.\\ :";
    char    DriveName[] = " :\\";
    HANDLE  Device;
    DWORD   BytesReturned;

    VolumeName[4] = DriveLetter;
    DriveName[0] = DriveLetter;

    Device = CreateFile(
        VolumeName,
        GENERIC_READ | GENERIC_WRITE,
        FILE_SHARE_READ | FILE_SHARE_WRITE,
        NULL,
        OPEN_EXISTING,
        FILE_FLAG_NO_BUFFERING,
        NULL
        );

    if (Device == INVALID_HANDLE_VALUE)
    {
        PrintLastError(&VolumeName[4]);
        return -1;
    }

    if (!DeviceIoControl(
        Device,
        FSCTL_LOCK_VOLUME,
        NULL,
        0,
        NULL,
        0,
        &BytesReturned,
        NULL
        ))
    {
        PrintLastError(&VolumeName[4]);
        CloseHandle(Device);
        return -1;
    }

    if (!DeviceIoControl(
        Device,
        IOCTL_FILE_DISK_CLOSE_FILE,
        NULL,
        0,
        NULL,
        0,
        &BytesReturned,
        NULL
        ))
    {
        PrintLastError("FileDisk:");
        CloseHandle(Device);
        return -1;
    }

    if (!DeviceIoControl(
        Device,
        FSCTL_DISMOUNT_VOLUME,
        NULL,
        0,
        NULL,
        0,
        &BytesReturned,
        NULL
        ))
    {
        PrintLastError(&VolumeName[4]);
        CloseHandle(Device);
        return -1;
    }

    if (!DeviceIoControl(
        Device,
        FSCTL_UNLOCK_VOLUME,
        NULL,
        0,
        NULL,
        0,
        &BytesReturned,
        NULL
        ))
    {
        PrintLastError(&VolumeName[4]);
        CloseHandle(Device);
        return -1;
    }

    CloseHandle(Device);

    if (!DefineDosDevice(
        DDD_REMOVE_DEFINITION,
        &VolumeName[4],
        NULL
        ))
    {
        PrintLastError(&VolumeName[4]);
        return -1;
    }

    SHChangeNotify(SHCNE_DRIVEREMOVED, SHCNF_PATH, DriveName, NULL);

    return 0;
}

int FileDiskStatus(char DriveLetter)
{
    char                    VolumeName[] = "\\\\.\\ :";
    HANDLE                  Device;
    POPEN_FILE_INFORMATION  OpenFileInformation;
    DWORD                   BytesReturned;

    VolumeName[4] = DriveLetter;

    Device = CreateFile(
        VolumeName,
        GENERIC_READ,
        FILE_SHARE_READ | FILE_SHARE_WRITE,
        NULL,
        OPEN_EXISTING,
        FILE_FLAG_NO_BUFFERING,
        NULL
        );

    if (Device == INVALID_HANDLE_VALUE)
    {
        PrintLastError(&VolumeName[4]);
        return -1;
    }

    OpenFileInformation = malloc(sizeof(OPEN_FILE_INFORMATION) + MAX_PATH);

    if (!DeviceIoControl(
        Device,
        IOCTL_FILE_DISK_QUERY_FILE,
        NULL,
        0,
        OpenFileInformation,
        sizeof(OPEN_FILE_INFORMATION) + MAX_PATH,
        &BytesReturned,
        NULL
        ))
    {
        PrintLastError(&VolumeName[4]);
        CloseHandle(Device);
        free(OpenFileInformation);
        return -1;
    }

    if (BytesReturned < sizeof(OPEN_FILE_INFORMATION))
    {
        SetLastError(ERROR_INSUFFICIENT_BUFFER);
        PrintLastError(&VolumeName[4]);
        CloseHandle(Device);
        free(OpenFileInformation);
        return -1;
    }

    CloseHandle(Device);

    printf("%c: %.*s %I64u bytes%s\n",
        DriveLetter,
        OpenFileInformation->FileNameLength,
        OpenFileInformation->FileName,
        OpenFileInformation->FileSize,
        OpenFileInformation->ReadOnly ? " ro" : ""
        );

    free(OpenFileInformation);

    return 0;
}

int __cdecl main(int argc, char* argv[])
{
    char*                   	Command;
	int							Pos;
	int							Count;
	int							Partition;
    int                     	DeviceNumber;
    char*                   	FileName;
    char*                   	Option;
    char                    	DriveLetter;
    BOOLEAN						CdImage = FALSE;
	PartitionEntry				Entries[16];
    POPEN_FILE_INFORMATION_EX	OpenFileInformation;

    Command = argv[1];

    if ((argc >= 5 && argc <= 7) && !strcmp(Command, "/mount"))
    {
        DeviceNumber = atoi(argv[2]);
        FileName = argv[3];
		Partition = 0;

        if (strlen(FileName) < 2)
        {
            return FileDiskSyntax();
        }

        OpenFileInformation =
            malloc(sizeof(OPEN_FILE_INFORMATION_EX) + strlen(FileName) + 7);

        memset(
            OpenFileInformation,
            0,
            sizeof(OPEN_FILE_INFORMATION_EX) + strlen(FileName) + 7
            );

        if (FileName[0] == '\\')
        {
            if (FileName[1] == '\\')
                // \\server\share\path\filedisk.img
            {
                strcpy(OpenFileInformation->FileName, "\\??\\UNC");
                strcat(OpenFileInformation->FileName, FileName + 1);
            }
            else
                // \Device\Harddisk0\Partition1\path\filedisk.img
            {
                strcpy(OpenFileInformation->FileName, FileName);
            }
        }
        else
            // c:\path\filedisk.img
        {
            strcpy(OpenFileInformation->FileName, "\\??\\");
            strcat(OpenFileInformation->FileName, FileName);
        }

        OpenFileInformation->FileNameLength =
            (USHORT) strlen(OpenFileInformation->FileName);

        for (Pos=4 ; Pos<argc-1; Pos++)
        {
            Option = argv[Pos];

            if (!strcmp(Option, "/ro"))
            {
                OpenFileInformation->ReadOnly = TRUE;
            }
            else if (!strcmp(Option, "/cd"))
            {
                CdImage = TRUE;
            }
            else if (Option[0]=='/' && Option[1]=='p')
            {
                Partition = atoi(Option+2);

				if(Partition>0)							// find partition in the file
				{
					Count = ScanMBR(FileName,Entries,16);

					if(Count<0)
					{
						PrintLastError(OpenFileInformation->FileName);
						return -1;
					}

					if(Count==0)
					{
						printf("\nno parition found !\n\n");
						return -1;
					}

					if(Count<=Partition)
					{
						printf("\nthe parition wasn't found in the file !\n\n");
						return -1;
					}


				OpenFileInformation->FileOffset.QuadPart = Entries[Partition-1].qwStart;
				OpenFileInformation->FileSize  .QuadPart = Entries[Partition-1].qwSize;
				}

            }
            else
            {
                if (Option[strlen(Option) - 1] == 'G')
                {
                    OpenFileInformation->FileSize.QuadPart =
                        _atoi64(Option) * 1024 * 1024 * 1024;
                }
                else if (Option[strlen(Option) - 1] == 'M')
                {
                    OpenFileInformation->FileSize.QuadPart =
                        _atoi64(Option) * 1024 * 1024;
                }
                else if (Option[strlen(Option) - 1] == 'k')
                {
                    OpenFileInformation->FileSize.QuadPart =
                        _atoi64(Option) * 1024;
                }
                else
                {
                    OpenFileInformation->FileSize.QuadPart =
                        _atoi64(Option);
                }
            }
        }

        DriveLetter = argv[Pos][0];
        OpenFileInformation->DriveLetter = DriveLetter;
        return FileDiskMount(DeviceNumber, OpenFileInformation, CdImage);
    }
    else if (argc == 3 && !strcmp(Command, "/list"))
    {
        return FileDiskListPartitions(argv[2]);
    }
    else if (argc == 3 && !strcmp(Command, "/umount"))
    {
        DriveLetter = argv[2][0];
        return FileDiskUmount(DriveLetter);
    }
    else if (argc == 3 && !strcmp(Command, "/status"))
    {
        DriveLetter = argv[2][0];
        return FileDiskStatus(DriveLetter);
    }
    else
    {
        return FileDiskSyntax();
    }
}
